/***************************************************************************
*                                                                          *
*	File :	vtf_lib.c                                                  *
*	Note :	Functions for a frequency-domain simulation of the vocal   *
*		tract with a nasal side branch. The vocal tract transfer   *
*		ratio, (U_lip + U_nostrils)/U_glottis, is calculated for   *
*		a specified frequency.					   *
*									   *
*		An extra fixed resistance is added to each of the nasal	   *
*		sections.  It affects on the appearance of the vowel	   *
*		spectral modification due to the occurence of pole-zero	   *
*		pairs.							   *
*									   *
*		The vocal tract is excited at the glottis in the current   *
*		program.  But series pressure or/and parallel current	   *
*		sources can be placed anywhere inside the vocal tract by   *
*		specifying eq[2j].s for a parallele current source and	   *
*		by eq[2j+1].s for a series pressure source.		   *
*									   *
***************************************************************************/

#include	"always.h"

#include	"complex.h"
#include	"plot_lib.h"
#include	"vtsimul.h"

/********************(structure array definitions)***********************/

typedef struct { double	Rs,	/* series resistance			*/
			Ls,	/* series inductance (acoustic mass)	*/
			Ca,	/* parallel capacitance (compliance)	*/
			G,	/* conductance due to heat conduction   */
			Rw,	/* wall mechanical resistance		*/
			Lw,	/* wall mass (inductance)		*/
			Cw;	/* wall compliance			*/
		}  fd_acoustic_elements;

typedef	struct { cmplx	s,	/* interlaced voltage & current sources	*/
			w,	/* interlaced admittances & impedances	*/
			x;	/* interlaced P's and U's to be solved	*/
		}  fd_linear_equation;


/*************************( global variables )*****************************/

	static	double	Rs, La, Ca, G, Rw, Lw, Cw, Rrad, Lrad;
/* pharyngeal tube */
	static	int			nph2, nph3, nph4;
	static	area_function		*afph;
	static	fd_acoustic_elements	*acph;
	static	fd_linear_equation	*eqph;
/* bucal tube	*/
	static	int			nbu2, nbu3, nbu4;
	static	area_function		*afbu;
	static	fd_acoustic_elements	*acbu;
	static	fd_linear_equation	*eqbu;
/* nasal tract */
	static	int			nna2, nna3, nna4;
	static	fd_acoustic_elements	*acna;
	static	fd_linear_equation	*eqna;
/* normalized radiation impedance */
	static	float	*rad_re, *rad_im;
	static	float	dka;
	static	int	n_ka;
	static	double	lip_Rrad, lip_Lrad;
	static	double	nose_Rrad, nose_Lrad;
	static	double	lip_rad_ka, lip_rad_norm;
	static	double	nose_rad_ka, nose_rad_norm;


/******************( Global variables defined in "main" )****************/

extern	int	nbu;		/* # of sections in the bucal tube	*/
extern	int	nph;		/* # of sections in the phryngeal tube	*/
extern	area_function *afvt;	/* AF from the glottis to the lips	*/
extern	double	anc;		/* nasal coupling area in cm2		*/
extern	int	nna;		/* # of sections in the nasal tract	*/
extern	area_function *afnt;	/* AF from nostrils to coupling point	*/

	/*( simulation options )*/

extern	int	nasal_tract;
extern	int	wall;
extern	int	rad_boundary;
extern	int	glt_boundary;
extern	int	source_loc;

	/*( an extra loss factor to increase heat loss in the nasal tract )*/

extern	double	extra_loss_factor;		/* in % */

	/*( physical constants )*/

extern	double	ro;		/* = 1.14e-3, air density, gm/cm**3	*/
extern	double	c;		/* = 3.5e+4, sound velocity, cm/s		*/
extern	double	eta;		/* = 1.4, adiabatic constant		*/
extern	double	cp;		/* = 0.24, specific heat, cal/gm.degree	*/
extern	double	lamda;		/* = 5.5e-5, heat conduction, cal/cm.sec.degree */
extern	double	mu;		/* = 1.86e-4, viscosity coef, dyne.sec/cm**2 */
extern	double	wall_resi;	/* = 1600., wall resistance, gm/s/cm2	*/
extern	double	wall_mass;	/* = 1.5, wall mass per area, gm/cm2	*/
extern	double	wall_comp;	/* = 3.0e+5, wall compliance		*/

extern	char	*RADIMPpath;


/***************************( Local functions )***************************/

/*****
*	Function: nonzero
*	Note	: limit x to a small nonzero positive value.
*****/
double	nonzero( double x )
{
	double	limit = 0.0001;

	if( x >= limit ) return( x );
	else             return( limit );
}


/*****
*	Function: copy_initial_af
*	Note :	Copy an initial input vocal tract area function, af,
*		having nss ( = nph + nbu) sections onto afph and afbu
*		with check zeros. Note that the address for the bucal
*		tube will be reversed, afbu1[0] corresponds to the lip
*		end and afbu1[nbu-1] to the nasal branch point.
****/

void	copy_initial_af ( void )
{
	int	i, j;

	for(i=0; i<nph; i++)
	{  afph[i].A = nonzero(afvt[i].A);
	   afph[i].x = nonzero(afvt[i].x);
	}
	for(i=0, j=nph+nbu-1; i<nbu; i++, j--)
	{  afbu[i].A = nonzero(afvt[j].A);
	   afbu[i].x = nonzero(afvt[j].x);
	}
}


/*****
*	Function: acou_elements
*	Note  :	compute coefficient values so that acoustic elements
*		of the transmission line can be calculated as a function
*		of a specified frequency value.
*****/
void	acou_elements (
	int			ns,		/* # of sections */
	area_function		af[],
	fd_acoustic_elements	ac[] )
{
	double	r0, r1, L0, L1, sqa, sqad, xda;
	int	i, j;

/* acoustic elements */

	r0 = 0;
	L0 = 0;

	for(i=0, j=1; i<ns; i++, j++)
	{  sqa  = sqrt(af[i].A);
	   sqad = sqa*af[i].x;
	   xda  = af[i].x / af[i].A;
	   r1   = Rs*xda/sqa;
	   L1   = La*xda;
	   ac[i].Rs = r0 + r1; r0 = r1;		/* series elements */
	   ac[i].Ls = L0 + L1; L0 = L1;
	   ac[j].Ca = Ca*af[i].A*af[i].x;	/* parallel elements */
	   ac[j].G  = G*sqad;
	}
	ac[ns].Rs = r1;			/* left arm of the last section	*/
	ac[ns].Ls = L1;

	if( wall == YIELDING )			/* yielding walls */
	   for(i=0, j=1; i<ns; i++, j++)
	   {  sqad     = af[i].x*sqrt(af[i].A);
	      ac[j].Rw = Rw/sqad;
	      ac[j].Lw = Lw/sqad;
	      ac[j].Cw = Cw*sqad;
	   }
}

/*****
*	Function: eq_elements
*	Note  :	compute the values of the force constants and coefficients
*		for the linear equation, at a given frequency.
*****/
void	eq_elements (
	double			freq,		/* frequency in Hz	*/
	int			ns,		/* number of sections	*/
	fd_acoustic_elements	ac[],
	fd_linear_equation	eq[]  )
{
	double	sq_freq;
	double	re, im, r, wl, wc, wlc, d;
	int	i, j;
	cmplx	a;

	sq_freq = sqrt(freq);

	a.re  = sq_freq*ac[0].Rs;		/* right arm */
	a.im  =    freq*ac[0].Ls;
	eq[1].w = a;

	for(i=1, j=1; i<=ns; i++)
	{  if( wall == YIELDING )
	   {  r   = ac[i].Rw;			/* parallel element */
	      wl  = freq*ac[i].Lw;
	      wc  = freq*ac[i].Cw;
	      wlc = wl - 1.0/wc;
	      d   = 1.0/(r*r + wlc*wlc);
	      a.re  = sq_freq*ac[i].G + r*d;
	      a.im  = freq*ac[i].Ca - wlc*d;
	   }
	   else
	   {  a.re  = sq_freq*ac[i].G;
	      a.im  = freq*ac[i].Ca;
	   }
	   eq[++j].w = a;

	   a.re  = sq_freq*ac[i].Rs;		/* series element */
	   a.im  =    freq*ac[i].Ls;
	   eq[++j].w = a;
	}
}


/******
*	Function: elimination
*	Note	: an in place forward elimination procedure to solve
*		  the linear algebraic equation, s = wx.
******/
void	elimination(
	int		ns3,	/* = 2*ns+1, ou ns = number of sections	*/
	fd_linear_equation  eq[] )

{
	int	i;
	cmplx	a;	/* complex valued "unity" */

	a.re = 1.;
	a.im = 0.;

	eq[1].w = cmplx_add_mul( a, eq[0].w, eq[1].w );
	eq[1].s = cmplx_add_mul( eq[0].s, eq[0].w, eq[1].s );

	for(i=2; i<=ns3; i++)
	{  eq[i].w = cmplx_add_mul( eq[i-2].w, eq[i-1].w, eq[i].w );
	   eq[i].s = cmplx_add_mul( eq[i-1].s, eq[i-1].w, eq[i].s );
	}
}

/******
*	Function: substitution
*	Note	: a backward substitution procedure to solve
*		  the linear algebraic equation, s = wx.
******/
void	substitution(
	int		ns3,	/* =2*ns+1, ou ns = number of sections	*/
	fd_linear_equation  eq[] )

{
	int	i;

	for(i=ns3; i>=1; i--)
	   eq[i].x = cmplx_div (
	      cmplx_sub_mul( eq[i].s, eq[i-1].w, eq[i+1].x ), eq[i].w
				);

	eq[0].x = cmplx_div( cmplx_sub( eq[0].s, eq[1].x ), eq[0].w );
}

/*****
*	Function : read_rad
*	Note :	read normalized radiation impedance from the current
*		directory.
*****/
/***  NOTE!!!! This had to be rewritten to deal with a text version
      of the data file (rad_imp.text) as the binary version 
      (rad_imp.dat) did not  work correctly in unix. -- FG, 7/24/96 ***/

void	read_rad (void)
{
	FILE	*in;
	int i, rad_re_pointer, rad_im_pointer;

	if ((in = fopen(RADIMPpath, "rb")) == NULL)
	  printf("Could not open input file %s\n",RADIMPpath);
	/* old maeda reads commented out
	fread(&dka, sizeof(float), 1, in);
	fread(&n_ka, sizeof(int), 1, in);
	*/
	fscanf(in,"%f",&dka);
	fscanf(in,"%d",&n_ka);
	rad_re = (float *) calloc(n_ka, sizeof(float));
	rad_im = (float *) calloc(n_ka, sizeof(float));
	for (i = 0; i < n_ka; i++)
	  {
	    fscanf(in,"%f %f",&rad_re[i],&rad_im[i]);
	  }
	
	/* old maeda reads commented out
	fread(rad_re, sizeof(float), n_ka, in);
	fread(rad_im, sizeof(float), n_ka, in);
	*/
	fclose(in);
}

/*****
*	Function : radimp
*	Note :	returns complex radiation impedance at a given freq..
*****/

cmplx	radimp (double ka)
{
	double	re, im, x1, x2, y1, y2;
	int	k;

	cmplx	a;

	if( ka >= dka*n_ka )
	{  a.re = rad_re[n_ka - 1];
	   a.im = rad_im[n_ka - 1];
	}
	else
	{  k = ka/dka;
	   if( k == 0 ) { x1 = 0; y1 = 0;}
	   if( k >= 1 ) { x1 = rad_re[k - 1]; y1 = rad_im[k - 1];}
	   x2 = rad_re[k];
	   y2 = rad_im[k];
	   a.re = x1 + (ka - k*dka)*(x2 - x1)/dka;
	   a.im = y1 + (ka - k*dka)*(y2 - y1)/dka;
	}
	return ( a );
}


/******************( Functions called from main )***********************/

/*****
*	Function : vtf_ini
*	Note :	Compute acoustic element coefficients from which their
*		values are determined as a function of the frequency (Hz).
*****/

void	vtf_ini ( void )
{
	double	omega_ini;
	int	i;
	double	pi = 3.141593;

	nph2 = 2*nph; nph3 = nph2+1; nph4 = nph2+2;
	nbu2 = 2*nbu; nbu3 = nbu2+1; nbu4 = nbu2+2;
	nna2 = 2*nna; nna3 = nna2+1; nna4 = nna2+2;

	/*** Coefficients for computing acoustic elements ***/
	omega_ini = 2.0*pi;
	
	/* series registance due to visscus loss */
	Rs  = sqrt(0.5*pi*omega_ini*ro*mu);

	/* acoustic elements */
	La = omega_ini*ro/2.0;	/* acoustic mass (La)		*/
	Ca = omega_ini/(ro*c*c);	/* acoustic stiffness (1/Ca)	*/

	/* parallel conductance due to heat conduction loss */
	G =  (eta-1.)*sqrt(pi*omega_ini*lamda/(ro*cp))/(ro*c*c);

	/* walls */
	Rw = wall_resi/(2.0*sqrt(pi));
	Lw = wall_mass*sqrt(pi);
	Cw = wall_comp*sqrt(pi);

	/* radiation impedance */
	Rrad = 2.0*pi*ro/c;
	Lrad = 16.0*ro/(3.0*sqrt(pi));

	/* read normalized radiation impedance from the file */
	if( rad_boundary == BESSEL_FUNCTION ) read_rad();

	/*** memory allocations ***/
	afph = (area_function *) calloc( nph, sizeof(area_function) );
	acph = (fd_acoustic_elements *) calloc( nph+1, 
						sizeof(fd_acoustic_elements));
	eqph = (fd_linear_equation *) calloc( 2*nph+3, 
					      sizeof(fd_linear_equation) );
	afbu = (area_function *) calloc( nbu, sizeof(area_function) );
	acbu = (fd_acoustic_elements *) calloc( nbu+1, 
						sizeof(fd_acoustic_elements) );
	eqbu = (fd_linear_equation *) calloc( 2*nbu+3, 
					      sizeof(fd_linear_equation) );
	acna = (fd_acoustic_elements *) calloc( nna+1, 
						sizeof(fd_acoustic_elements) );
	eqna = (fd_linear_equation *) calloc( 2*nna+3, 
					      sizeof(fd_linear_equation) );

	/**** Acoustic and matrix elements ****/
	copy_initial_af();		/* copy the VT area function */

	/* pharyngeal tract */
	acou_elements( nph, afph, acph );

	/* bucal cavity */
	acou_elements( nbu, afbu, acbu );

	lip_Rrad = Rrad;
	lip_Lrad = Lrad/sqrt(afbu[0].A);

	lip_rad_ka = 2.0*sqrt(pi*afbu[0].A)/c;
	lip_rad_norm = ro*c/afbu[0].A;

	/* nasal tract */
	if( nasal_tract == ON )
	{  afnt[nna-1].A = nonzero( anc );	/* nasal coupling */
	   acou_elements( nna, afnt, acna );
	   for(i=0; i<=nna; i++)
	      acna[i].G *= extra_loss_factor;	/* ad.hoc.extra loss */

	   nose_Rrad = Rrad;
	   nose_Lrad = Lrad/sqrt(afnt[0].A);

	   nose_rad_ka = 2.0*sqrt(pi*afnt[0].A)/c;
	   nose_rad_norm = ro*c/afnt[0].A;
	}
}


/*****
*	Function: vtf_sim
*	Note :	compute vocal tract transfer, U/Ug, at a specified
*		frequency (in Hz) in the frequency domain simulation.
*		Returns the transfer ratio.
*****/

double	vtf_sim( double freq )
{
	int	i, j;
	cmplx	f, g, h, p, q;
	cmplx	Uout;

	cmplx	a0;	/* complex valued "zero" */
	cmplx	a1;	/* complex valued "unity" */

	a0.re = 0.;
	a0.im = 0.;
	a1.re = 1.;
	a1.im = 0.;

/*** pharyngeal cavity ***/

	eq_elements( freq, nph, acph, eqph );

/* glottal boundary condition, open or close */
	if(source_loc == 0) glt_boundary = CLOSE;	/* by default */
	if(glt_boundary == CLOSE) eqph[0].w = a0;
	else			  eqph[0].w = complex( 1.0e+8, -1.0e+8 );

/*** bucal cavity ***/

	eq_elements( freq, nbu, acbu, eqbu );

/* lip end boundary condition */
	switch( rad_boundary )
	{
	case BESSEL_FUNCTION :			/* circular piston set on */
	   eqbu[0].w = cmplx_inv (		/* an infinite baffle.	  */
	      cmplx_scl( lip_rad_norm, radimp(freq*lip_rad_ka ) ) );
	   break;

	case RL_CIRCUIT :			/* R-L circuit approximation */
	   eqbu[0].w = cmplx_inv (
	      complex( freq*freq*lip_Rrad, freq*lip_Lrad ) );
	   break;

	case SHORT_CIRCUIT :			/* nearly short circuit */
	   eqbu[0].w = complex(1.0e+8, -1.0e+8);
	   break;
	}

/*** Unit source inside the pharyngeal and bucal cavities ***/

	for(i=0; i<=nph3; i++) eqph[i].s = a0;
	for(i=0; i<=nbu3; i++) eqbu[i].s = a0;

	if(source_loc <= nph)
	{  if(source_loc == 0)  eqph[0].s = a1;	/* flow */
	   else  eqph[2*source_loc+1].s = complex( 10., 0. );	/* pressure */
	}
	else
	{  j = (nph+nbu) - source_loc;
	   eqbu[2*j+1].s = complex(100., 0. );
	}

/*** nasal tract ***/

	if( nasal_tract == ON )
	{  eq_elements( freq, nna, acna, eqna );
	   for(i=0; i<=nna3; i++) eqna[i].s = a0;

	/* nostrils boundary condition */

	   switch( rad_boundary )
	   {
	   case BESSEL_FUNCTION :
	      eqna[0].w = cmplx_inv (
		 cmplx_scl( nose_rad_norm, radimp(freq*nose_rad_ka) ) );
	      break;

	   case RL_CIRCUIT :
	      eqna[0].w = cmplx_inv (
		 complex( freq*freq*nose_Rrad, freq*nose_Lrad ) );
	      break;

	   case SHORT_CIRCUIT :
	      eqna[0].w = complex(1.0e+8, -1.0e+8);
	      break;
	   }
	}

/* solve s = wx */
	if( nasal_tract	== ON )
	{  elimination(nph3, eqph);
	   elimination(nbu3, eqbu);
	   elimination(nna3, eqna);

	   f = cmplx_div( eqph[nph3].s, eqph[nph3].w );
	   g = cmplx_div( eqbu[nbu3].s, eqbu[nbu3].w );
	   h = cmplx_div( eqna[nna3].s, eqna[nna3].w );
	   p = cmplx_add( f, cmplx_add( g, h ) );

	   f = cmplx_div( eqph[nph2].w, eqph[nph3].w );
	   g = cmplx_div( eqbu[nbu2].w, eqbu[nbu3].w );
	   h = cmplx_div( eqna[nna2].w, eqna[nna3].w );
	   q = cmplx_add( f, cmplx_add( g, h ) );

	   eqph[nph4].x = eqbu[nbu4].x = eqna[nna4].x = cmplx_div( p, q );

	 /*substitution(nph3, eqph);*/
	   substitution(nbu3, eqbu);
	   substitution(nna3, eqna);
	}
	else
	{  elimination(nph3, eqph);
	   elimination(nbu3, eqbu);

	   f = cmplx_div( eqph[nph3].s, eqph[nph3].w );
	   g = cmplx_div( eqbu[nbu3].s, eqbu[nbu3].w );
	   p = cmplx_add( f, g );

	   f = cmplx_div( eqph[nph2].w, eqph[nph3].w );
	   g = cmplx_div( eqbu[nbu2].w, eqbu[nbu3].w );
	   q = cmplx_add( f, g );

	   eqph[nph4].x = eqbu[nbu4].x = cmplx_div( p, q );

	 /*substitution(nph3, eqph);*/
	   substitution(nbu3, eqbu);
	}

/*** return the transfer ratio ***/

	Uout = eqbu[1].x;
	if( nasal_tract == ON ) Uout = cmplx_add( Uout, eqna[1].x );
	return( abs_cmplx( Uout) );
}

/*****
*	Function: vtf_UPdistribution
*	Note :	compute volume velocity and pressure distribution inside
*		the vocal tract for a given frequency (in Hz), in the
*		frequency domain simulation.
*****/

void	vtf_UPdistribution (
	double	freq,		/* frequency in Hz */
	double	*U,		/* volume velocity distribution */
	double	*P )		/* presure distribution */
{
	int	i, j;
	cmplx	f, g, h, p, q;



/*** pharyngeal cavity ***/

	eq_elements( freq, nph, acph, eqph );

/* glottal boundary condition, open or close */
	if(source_loc == 0) glt_boundary = CLOSE;	/* by default */
	if(glt_boundary == CLOSE) eqph[0].w = complex( 0., 0.);
	else			  eqph[0].w = complex( 1.0e+8, -1.0e+8 );

/*** bucal cavity ***/

	eq_elements( freq, nbu, acbu, eqbu );

/* lip end boundary condition */
	switch( rad_boundary )
	{
	case BESSEL_FUNCTION :			/* circular piston set on */
	   eqbu[0].w = cmplx_inv (		/* an infinite baffle.	  */
	      cmplx_scl( lip_rad_norm, radimp(freq*lip_rad_ka ) ) );
	   break;

	case RL_CIRCUIT :			/* R-L circuit approximation */
	   eqbu[0].w = cmplx_inv (
	      complex( freq*freq*lip_Rrad, freq*lip_Lrad ) );
	   break;

	case SHORT_CIRCUIT :			/* nearly short circuit */
	   eqbu[0].w = complex(1.0e+8, -1.0e+8);
	   break;
	}

/*** Unit source inside the pharyngeal and bucal cavities ***/

	for(i=0; i<=nph3; i++) eqph[i].s = complex( 0., 0.);
	for(i=0; i<=nbu3; i++) eqbu[i].s = complex( 0., 0.);

	if(source_loc <= nph)
	{  if(source_loc == 0)  eqph[0].s = complex( 1., 0. );	/* flow */
	   else  eqph[2*source_loc+1].s = complex( 10., 0. );	/* pressure */
	}
	else
	{  j = (nph+nbu) - source_loc;
	   eqbu[2*j+1].s = complex(10., 0. );
	}

/*** nasal tract ***/

	if( nasal_tract == ON )
	{  eq_elements( freq, nna, acna, eqna );
	   for(i=0; i<=nna3; i++) eqna[i].s = complex( 0., 0.);

	/* nostrils boundary condition */

	   switch( rad_boundary )
	   {
	   case BESSEL_FUNCTION :
	      eqna[0].w = cmplx_inv (
		 cmplx_scl( nose_rad_norm, radimp(freq*nose_rad_ka) ) );
	      break;

	   case RL_CIRCUIT :
	      eqna[0].w = cmplx_inv (
		 complex( freq*freq*nose_Rrad, freq*nose_Lrad ) );
	      break;

	   case SHORT_CIRCUIT :
	      eqna[0].w = complex(1.0e+8, -1.0e+8);
	      break;
	   }
	}

/* solve s = wx */
	if( nasal_tract	== ON )
	{  elimination(nph3, eqph);
	   elimination(nbu3, eqbu);
	   elimination(nna3, eqna);

	   f = cmplx_div( eqph[nph3].s, eqph[nph3].w );
	   g = cmplx_div( eqbu[nbu3].s, eqbu[nbu3].w );
	   h = cmplx_div( eqna[nna3].s, eqna[nna3].w );
	   p = cmplx_add( f, cmplx_add( g, h ) );

	   f = cmplx_div( eqph[nph2].w, eqph[nph3].w );
	   g = cmplx_div( eqbu[nbu2].w, eqbu[nbu3].w );
	   h = cmplx_div( eqna[nna2].w, eqna[nna3].w );
	   q = cmplx_add( f, cmplx_add( g, h ) );

	   eqph[nph4].x = eqbu[nbu4].x = eqna[nna4].x = cmplx_div( p, q );

	   substitution(nph3, eqph);
	   substitution(nbu3, eqbu);
	   substitution(nna3, eqna);
	}
	else
	{  elimination(nph3, eqph);
	   elimination(nbu3, eqbu);

	   f = cmplx_div( eqph[nph3].s, eqph[nph3].w );
	   g = cmplx_div( eqbu[nbu3].s, eqbu[nbu3].w );
	   p = cmplx_add( f, g );

	   f = cmplx_div( eqph[nph2].w, eqph[nph3].w );
	   g = cmplx_div( eqbu[nbu2].w, eqbu[nbu3].w );
	   q = cmplx_add( f, g );

	   eqph[nph4].x = eqbu[nbu4].x = cmplx_div( p, q );

	   substitution(nph3, eqph);
	   substitution(nbu3, eqbu);
	}

/*** distibutions ***/

	for(i=0; i<=nph; i++)
	{  U[i-1] = abs_cmplx( eqph[2*i+1].x );
	   P[i-1] = abs_cmplx( eqph[2*i].x );
	}
	for(i=0; i<=nbu; i++)
	{  U[nph+nbu-i] = abs_cmplx( eqbu[2*i-1].x );
	   P[nph+nbu-i] = abs_cmplx( eqbu[2*i].x );
	}
}

/*****
*	Function : vtf_term
*	Note :	free memories
****/

void	vtf_term ( void )
{
	free( afph );
	free( acph );
	free( eqph );

	free( afbu );
	free( acbu );
	free( eqbu );

	free( acna );
	free( eqna );

	if( rad_boundary == BESSEL_FUNCTION )
	{  free( rad_re );
	   free( rad_im );
	}
}



#ifdef COMMENT
/******
*	Function : calplot_tf_FA
*	Note :	Compute Ul/Ug transfer ratio of a given VT area function
*		and plot.  Formant frequencies and amplitudes are
*		estimated by a peak picking algorithm with a variable
*		frequency step size scheme.  The transfer is plotted
*		but not returned to calling program.
*****/
void	calplot_tf_FA (
	char	*title,		/* a main title for the graphics frame */
	int	nfrmmax,	/* maximum number of F-peaks to be detected */
	double	*frm,		/* F1, F2, .... frequency in Hz */
	double	*amp,		/* A1, A2, .... amplitude in dB */
	int	*nfrms,		/* number of F-peaks detected   */
	vp_lo_frame  vp )
{
	double	df, dfmin = .5, dfmax = 40.;			/* in Hz */
	double	freq, freq2, mag, mag1, mag2, mag3;
	double	fmin = 0., ftic = 0.2, fmax = 5.;		/* in kHz */
	int	i, j;


	def_vp (vp.n, vp.x, vp.y, vp.x+vp.w, vp.y+vp.h, 1);
	open_vp(vp.n);
	clear_vp(vp.n);
	g_frame ( vp.n,
		 1, title, "Frequency (kHz)","Relative mag. (dB)",
		 fmin,  40.,  fmax, -40.0,
		 0, 5, fmin,  ftic, fmax,
		 0, 1, -40.0,  10.0, 40.0 );

	j  = 0;
	df = dfmax;

	vtf_ini();
	for( freq=df, i=0; freq <= 1000.*fmax; freq=freq+df, i++ )
	{  mag = 20.*log10( max(vtf_sim(freq), 0.01) );
	   if( i == 0 )
	   {  mov( datolo_x(vp.n, freq/1000.), datolo_y(vp.n, mag) );
	      mag1 = mag;
	   }
	   if( i == 1 )
	   {  drw( datolo_x(vp.n, freq/1000.), datolo_y(vp.n, mag) );
	      mag2 = mag;
	      freq2 = freq;
	   }
	   if( i > 1 )
	   {  drw( datolo_x(vp.n, freq/1000.), datolo_y(vp.n, mag) );
	      mag3 = mag;
/* detect spectral peak */
	      if( mag1<mag2 && mag2>=mag3 && j<nfrmmax )
	      {  frm[j]   = freq2;		/* Fn in Hz    */
		 amp[j++] = mag2;		/* An in dB    */
	      }
/* adapt frequency step size */
	      if ( abs(mag3 - mag2) > 1. )
	      {  df = df - 8;
		 if( df <= dfmin ) df = dfmin;
	      }
	      else
	      {  df = df + 8;
		 if( df >= dfmax ) df = dfmax;
	      }
	      mag1 = mag2;
	      mag2 = mag3;
	      freq2 = freq;
	   }
	}
	*nfrms = j;
	vtf_term();
}


/******
*	Function : calplot_tf_FBA
*	Note :	Compute Ul/Ug transfer ratio of a given VT area function
*		and plot.  Formant frrequencies and amplitudes are
*		estimated by a peak picking algorithm with a variable
*		frequency step size scheme.  The transfer is plotted
*		but not returned to calling program.
*	Modif.: David Gesbert added bandwidth detection routine (16/09/92).
*	Modif.: Maeda modified it with a binary search argorithm (23/09/92).
*****/
void	calplot_tf_FBA (
	char	*title,		/* a main title for the graphics frame */
	int	nfrmmax,	/* maximum number of F-peaks to be detected */
	double	*frm,		/* F1, F2, .... frequency in Hz */
	double	*bw,		/* B1, B2, .... frequency in Hz */
	double	*amp,		/* A1, A2, .... amplitude in dB */
	int	*nfrms,		/* number of F-peaks detected   */
	vp_lo_frame  vp )
{
	double	df, dfmin = .5, dfmax = 40.;			/* in Hz */
	double	freq, freq2, mag, mag1, mag2, mag3;
	double	fmin = 0., ftic = 0.2, fmax = 5.;		/* in kHz */
	int	i, j;

	/* variables used in David' algorithm */
/****
	int	change_direction, bw_not_found;
	double	pas, mag0;
****/
	/* variables used in binary search algorithm */

	double	dA = 0.5;					/* in dB */
	double	A3dB, left_half_bw, right_half_bw;

/* Graphics frame initialization */

	def_vp (vp.n, vp.x, vp.y, vp.x+vp.w, vp.y+vp.h, 1);
	open_vp(vp.n);
	clear_vp(vp.n);
	g_frame ( vp.n,
		 1, title, "Frequency (kHz)","Relative mag. (dB)",
		 fmin,  40.,  fmax, -40.0,
		 0, 5, fmin,  ftic, fmax,
		 0, 1, -40.0,  10.0, 40.0 );

/* Transfer ratio calculation */

	j  = 0;
	df = dfmax;
	vtf_ini();
	for( freq=df, i=0; freq <= 1000.*fmax; freq=freq+df, i++ )
	{  mag = 20.*log10( max(vtf_sim(freq), 0.01) );
	   if( i == 0 )
	   {  mov( datolo_x(vp.n, freq/1000.), datolo_y(vp.n, mag) );
	      mag1 = mag;
	   }
	   if( i == 1 )
	   {  drw( datolo_x(vp.n, freq/1000.), datolo_y(vp.n, mag) );
	      mag2 = mag;
	      freq2 = freq;
	   }
	   if( i > 1 )
	   {  drw( datolo_x(vp.n, freq/1000.), datolo_y(vp.n, mag) );
	      mag3 = mag;
/* detect spectral peak */
	      if( mag1<mag2 && mag2>=mag3 && j<nfrmmax )
	      {  frm[j]   = freq2;		/* Fn in Hz    */
		 amp[j++] = mag2;		/* An in dB    */
	      }
/* adapt frequency step size */
	      if ( abs(mag3 - mag2) > 1. )
	      {  df = df - 8;
		 if( df <= dfmin ) df = dfmin;
	      }
	      else
	      {  df = df + 8;
		 if( df >= dfmax ) df = dfmax;
	      }
	      mag1 = mag2;
	      mag2 = mag3;
	      freq2 = freq;
	   }
	}
	*nfrms = j;

/****** -3db formant bandwidth detection (David Gesbert) ******/
/****
	for(j=0;j< *nfrms;j++)
	{
	   change_direction=0;
	   pas=-4*dfmin;
	   bw_not_found=1;

	   do
	   {
	      if(change_direction) pas=-pas;
	      freq=frm[j];
	      mag0=20*log10(max(vtf_sim(freq),0.01));
	      mag=mag0;

	      do
	      {
		 freq+=pas;
		 mag1=mag;
		 mag=20*log10(max(vtf_sim(freq),0.01));
		 if(mag<mag0-3) {bw_not_found=0; break;}
	      }
	      while(((mag1>=mag)&&(fabs(freq-frm[j])>20)) ||
		     ((mag<mag0+0.1) && (fabs(freq-frm[j])<=20)));

	      if (change_direction && bw_not_found) {freq=frm[j]; break;}
	      change_direction=1;
	   }
	   while(bw_not_found);

	   bw[j]=2*fabs(freq-frm[j]);
	}
****/

/*** -3dB bandwidth detection with binary search algorithm ***/

	for(j=0; j<*nfrms; j++)
	{
	   A3dB = amp[j] - 3.0;

	/* -3dB point on the left skirt of the (j+1)-th formant */

	   if( j==0 ) df = frm[j]/2.;
	   else	 df = (frm[j] - frm[j-1])/2.;

	   freq = frm[j] - df;
	   mag = 20*log10(max(vtf_sim(freq),0.01));
	   if( mag > A3dB ) left_half_bw = 0;
	   else
	   {  for(i=0; i<20; i++)
	      {  df /= 2;
		 if( df<dfmin ) break;
		 mag = 20*log10(max(vtf_sim(freq),0.01));

		 if( mag > A3dB+dA )	  freq -= df;
		 else if( mag < A3dB-dA ) freq += df;
		 else			  break;
	      }
	      left_half_bw = frm[j] - freq;
	   }

	/* -3dB point on the right skirt of the (j+1)-th formant */

	   if( j==*nfrms-1 ) df = (1000.*fmax - frm[j])/2.;
	   else		     df = (frm[j+1] - frm[j])/2.;

	   freq = frm[j] + df;
	   mag = 20*log10(max(vtf_sim(freq),0.01));
	   if( mag > A3dB ) right_half_bw = 0;
	   else
	   {  for(i=0; i<20; i++)
	      {  df /= 2;
		 if( df<dfmin ) break;
		 mag = 20*log10(max(vtf_sim(freq),0.01));

		 if( mag > A3dB+dA )	  freq += df;
		 else if( mag < A3dB-dA ) freq -= df;
		 else			  break;
	      }
	      right_half_bw = freq - frm[j];
	   }
	   if( left_half_bw == 0. )  right_half_bw *= 2.;
	   if( right_half_bw == 0. ) left_half_bw  *= 2.;
	   bw[j] = left_half_bw + right_half_bw;
	}

	vtf_term();
}
#endif COMMENT



/******
*	Function : cal_tf_FA
*	Note :	Compute Ul/Ug transfer ratio of a given VT area function
*		and plot.  Formant frequencies and amplitudes are
*		estimated by a peak picking algorithm with a variable
*		frequency step size scheme.  The transfer function is
*		returned to calling program.
*
*	At the time this function is called, afvt[] must be set correctly.
*****/
void	cal_tf_FA (
	int	nfrmmax,	/* maximum number of F-peaks to be detected */
	double	*frm,		/* F1, F2, .... frequency in Hz */
	double	*amp,		/* A1, A2, .... amplitude in dB */
	int	*nfrms)		/* number of F-peaks detected   */
{
	double	df, dfmin = 0.5, dfmax = 20.;			/* in Hz */
	double	freq, freq2, mag, mag1, mag2, mag3;
	double	fmin = 0., ftic = 0.2, fmax = 5.;		/* in kHz */
	int	i, j;

	int	count=0;

	j  = 0;
	df = dfmax;
	vtf_ini();

	for( freq=df,i=0; (freq<=1000.*fmax) && (j<nfrmmax); freq=freq+df, i++ )
	{
	   count++;
	   mag = 20.*log10( max(vtf_sim(freq), 0.01) );
	   if( i == 0 )
	   {  /* mov( datolo_x(vp.n, freq/1000.), datolo_y(vp.n, mag) ); */
	      mag1 = mag;
	   }
	   if( i == 1 )
	   {  /* drw( datolo_x(vp.n, freq/1000.), datolo_y(vp.n, mag) ); */
	      mag2 = mag;
	      freq2 = freq;
	   }
	   if( i > 1 )
	   {  /* drw( datolo_x(vp.n, freq/1000.), datolo_y(vp.n, mag) ); */
	      mag3 = mag;
	      /* detect spectral peak */
	      if( mag1<mag2 && mag2>=mag3 )
	      {  frm[j]   = freq2;		/* Fn in Hz    */
		 amp[j++] = mag2;		/* An in dB    */
	      }
	      /* adapt frequency step size */
	      if ( abs(mag3 - mag2) > 1. )
	      {  df = df - 10;
		 if( df <= dfmin ) df = dfmin;
	      }
	      else
	      {  df = df + 10;
		 if( df >= dfmax ) df = dfmax;
	      }
	      mag1 = mag2;
	      mag2 = mag3;
	      freq2 = freq;
	   }
	}
	*nfrms = j;
	vtf_term();
}


