/*
 *      SenSyn - Version 1.0
 *
 *      Copyright (c) 1991, 1992 by Sensimetrics Corporation
 *      All Rights Reserved.
 *
 */

/* 
 *  frame.c - synthesize one frame
 *
 *  coded by Eric P. Carlson
 *
 *  Modification History:
 *     5 Nov 1991	removed sample.c and init_syn.c, renamed to frame.c
 *    21 Nov 1991	moved spectral_tilt, formant_1s to be pitch-synchronous
 *    27 Feb 1992	update formant_1s here, but DF1/DB1 pitch-synchronously
 *    26 Aug 1992	changed to take INT16 pointer as parameter
 */

#include <math.h>		/* needed for power and fabs function */
#include <limits.h>		/* for LONG_MAX and LONG_MIN */
#include "reson.h"		/* resonator support */
#include "synth.h"		/* synthesizer support */

extern long sigmx;

/* ---------------------------------------------------------------------- */

void senswave(INT16 *wave) {
  static Coefficients coefs;
  static int initialized = 0;
  int index;
  Float sample;

  /* initialize synth constants */
  if (!initialized) {
    init_synthesizer();
    initialized = 1;
  }

  /* setup amplitudes */
  coefs.asp_amp = pars[P_AH]?dB2amp(spkrdef[S_GH] + pars[P_AH] + A_AH):0;
  coefs.fric_amp = pars[P_AF]?dB2amp(spkrdef[S_GF] + pars[P_AF] + A_AF):0;
  coefs.f1p_amp = dB2amp(A_A1);
  coefs.f2p_amp = dB2amp(pars[P_A2F] + A_A2F);
  coefs.f3p_amp = dB2amp(pars[P_A3F] + A_A3F);
  coefs.f4p_amp = dB2amp(pars[P_A4F] + A_A4F);
  coefs.f5p_amp = dB2amp(pars[P_A5F] + A_A5F);
  coefs.f6p_amp = dB2amp(pars[P_A6F] + A_A6F);
  coefs.bypass_amp = dB2amp(pars[P_AB] + A_AB);
  coefs.npv_amp = dB2amp(pars[P_ANV] + A_ANV);
  coefs.f1v_amp = dB2amp(pars[P_A1V] + A_A1V);
  coefs.f2v_amp = dB2amp(pars[P_A2V] + A_A2V);
  coefs.f3v_amp = dB2amp(pars[P_A3V] + A_A3V);
  coefs.f4v_amp = dB2amp(pars[P_A4V] + A_A4V);
  coefs.tpv_amp = dB2amp(pars[P_ATV] + A_ATV);

  /* setup resonators */
  /* formant_1_cascade is also setup pitch-synchronously in voice.c */
  InterPolePair(&formant_1_cascade,
    pars[P_F1] + (synth.glottis_open ? pars[P_DF1] : 0),
    pars[P_B1] + (synth.glottis_open ? pars[P_DB1] : 0),
    spkrdef[S_SR]);
  InterPolePair(&formant_2_cascade, pars[P_F2], pars[P_B2], spkrdef[S_SR]);
  InterPolePair(&formant_3_cascade, pars[P_F3], pars[P_B3], spkrdef[S_SR]);
  SetPolePair(&formant_4_cascade, pars[P_F4], pars[P_B4], spkrdef[S_SR]);
  SetPolePair(&formant_5_cascade, pars[P_F5], pars[P_B5], spkrdef[S_SR]);
  SetPolePair(&formant_6_cascade, pars[P_F6], pars[P_B6], spkrdef[S_SR]);
  SetPolePair(&formant_7_cascade, 6500, 500, spkrdef[S_SR]);
  SetPolePair(&formant_8_cascade, 7500, 600, spkrdef[S_SR]);
  SetPolePair(&nasal_pole_cascade, pars[P_FNP], pars[P_BNP], spkrdef[S_SR]);
  SetZeroPair(&nasal_zero_cascade, pars[P_FNZ], pars[P_BNZ], spkrdef[S_SR]);
  SetPolePair(&trach_pole_cascade, pars[P_FTP], pars[P_BTP], spkrdef[S_SR]);
  SetZeroPair(&trach_zero_cascade, pars[P_FTZ], pars[P_BTZ], spkrdef[S_SR]);
  InterPolePair(&formant_2_parallel, pars[P_F2], pars[P_B2F], spkrdef[S_SR]);
  InterPolePair(&formant_3_parallel, pars[P_F3], pars[P_B3F], spkrdef[S_SR]);
  SetPolePair(&formant_4_parallel, pars[P_F4], pars[P_B4F], spkrdef[S_SR]);
  SetPolePair(&formant_5_parallel, pars[P_F5], pars[P_B5F], spkrdef[S_SR]);
  SetPolePair(&formant_6_parallel, pars[P_F6], pars[P_B6F], spkrdef[S_SR]);
  SetPolePair(&nasal_pole_special, pars[P_FNP], pars[P_BNP], spkrdef[S_SR]);
  /* formant_1_special is also setup pitch-synchronously in voice.c */
  InterPolePair(&formant_1_special,
    pars[P_F1] + (synth.glottis_open ? pars[P_DF1] : 0),
    pars[P_B1] + (synth.glottis_open ? pars[P_DB1] : 0),
    spkrdef[S_SR]);
  InterPolePair(&formant_2_special, pars[P_F2], pars[P_B2], spkrdef[S_SR]);
  InterPolePair(&formant_3_special, pars[P_F3], pars[P_B3], spkrdef[S_SR]);
  SetPolePair(&formant_4_special, pars[P_F4], pars[P_B4], spkrdef[S_SR]);
  SetPolePair(&trach_pole_special, pars[P_FTP], pars[P_BTP], spkrdef[S_SR]);

  /* setup noise source */
  if (spkrdef[S_SB] && !pars[P_AF] && !pars[P_AH])
    synth.random = spkrdef[S_RS];

  /* synthesize some samples */
  for (index = 0; index < spkrdef[S_UI]; index++) {
    sample = next_sample(&coefs);
    if (sample <= 32767)
      if (sample >= -32768)
	wave[index] = (INT16) (sample + 0.5);	/* round, not truncate */
      else
	wave[index] = -32768;
    else
      wave[index] = 32767;

    if (sample > (Float) LONG_MAX)
      sample = (Float) LONG_MAX;
    else if (sample < (Float) LONG_MIN)
      sample = (Float) LONG_MIN;

    if (labs((long) sample) > sigmx)
      sigmx = labs((long) sample);
  }
}
